#include <fstream>
#include <iostream>
#include <iomanip>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


using namespace std;

struct data
{
    double Alfa;
    double CL;
    double CD;
    double CDp;
    double CM;
    double Top_Xtr;
    double Bot_Xtr;
};      

double Gradient(int i, int n, int x, double *X, double F);
double Xfoil(double *X);

int Direction(ofstream &log, int i, int n, double *X, double F, double *dF, double *p);
int LambdaSearch(ofstream &log, int i, int n, double Lamb, double AlfaLim, double *X, double &F, double *dF, double *p, data &aero);
int Stop(ofstream &log, int n, double Fprev, double F, double Sum_dF, double dF_Stop, double F_Stop);
int Results(ofstream &log, ofstream &res, int i, int n, double &Sum_dF, double *X, double F, double *dF, data aero);

int XfoilData(double &a, double &CL, double &CD, double &CDp, double &CM, double &Top_Xtr, double &Bot_Xtr);

int main()
{ 
    //*** FILES INITIALIZATION ***
    
    ofstream res("Results.txt");

    if(!res)
    {
        cout << "Results.txt error!!!" << endl;
        return 0;
    }
    
    ofstream log("Log.txt");

    if(!log)
    {
        cout << "Log.txt error!!!" << endl;
        return 0;
    }
    //*** VARIABLES INITIALIZATION ***
    
    int Iter, n, AlfaIterLim;
    double F, Fprev, Lamb, Sum_dF;
    double dF_Stop, F_Stop;
    double X[3], dF[3], p[3];
    data aero;
    
    Iter = 30;
    
    AlfaIterLim = 10;
    Lamb = 0.005;
    
    dF_Stop = 0.000001;
    F_Stop  = 0.000001;
    
    n = 4;
    
    X[0] = 0.12;      //t
    X[1] = 0.0184;    //f
    X[2] = 0.3;       //x_t
    X[3] = 0.149;     //x_f
    
    //*** THIRST SOLUTION ***     
    
    res << "Iter\t";
    
    for(int x=0; x<n; x++) 
        res << "X" << x+1 << "\t";
    
    res << "F\t" << "1/F\t" << "Sum_dF\t\t"; 
    res << "Alfa\t" << "CL\t" << "CD\t" << "CDp\t" << "CM\t" << "Top_xtr\t" << "Bot_xtr" << endl;
    
    res << 0 << "\t" << fixed << setprecision(4) << X[0] << "\t";
        
    for(int x=1; x<n; x++)    
        res << X[x] << "\t";
           
    Fprev = F = Xfoil(X); 
  
    res << F << "\t" << 1/F << endl;
    
    //*** MAIN ITERATION LOOP ***
    
    for(int i=1; i<=Iter; i++)       
    { 
        log << endl << "ITER:\t" << i << endl;
        log << "---------------------------------------------" << endl << endl;
         
        //*** DIRECTION ESTIMATION *** 
           
        Direction(log, i, n, X, F, dF, p);
        
        //*** ALFA SEARCH ***  
        
        LambdaSearch(log, i, n, Lamb, AlfaIterLim, X, F, dF, p, aero);
        
        //*** RESULTS ***
        
        Results(log, res, i, n, Sum_dF, X, F, dF, aero);
        
        //*** STOP CRYTERIUM *** 
        
        int stop = 1;
        
        stop = Stop(log, n, Fprev, F, Sum_dF, dF_Stop, F_Stop);
        
        if(stop == 0)
            break;
        
        Fprev = F;
    }
    
    log.close();
    res.close();
    
    system("@del /q XfoilData.dat\n");
    system("PAUSE");
    return EXIT_SUCCESS;
}

double Xfoil(double *X)
{   
    //*** XFoil SCRIPT DEFINITION *** 
    
    ofstream out("XFoil.txt");

    if(!out)
    {
        cout << "XFoil.txt error!!!" << endl;
        return 0;
    }

    out << "NACA 23012" << endl;
    
    out << "GDES" << endl;
    out << "TSET" << endl;
    out << X[0] << endl;
    out << X[1] << endl;
    out << "HIGH" << endl;
    out << X[2] << endl;
    out << X[3] << endl;
    out << "CADD" << endl << endl << endl << endl;
    out << "EXEC" << endl << endl;
    
    out << "PANE" << endl;
    
    out << "MDES" << endl;
    out << "FILT" << endl;
    out << "EXEC" << endl << endl;
    
    out << "OPER" << endl;
    
    out << "VPAR" << endl;
    out << "N " << 9 << endl;
    out << "VACC " << 0.001 << endl << endl;
    
    out << "a " << 3 << endl;
    out << "VISC " << 200000 << endl;
    out << "ITER" << endl;
    out << 40 << endl;
    out << "a " << 3 << endl;
    out << "INIT" << endl;
    out << "a " << 3 << endl;
    out << "PACC" << endl;
    out << "XfoilData.dat" << endl << endl;
    out << "a " << 3 << endl;
    out << "PACC" << endl << endl;

    out << "QUIT" << endl;

    out.close();

    // *** XFoil COMPUTATION ***
    
    system("@if exist XfoilData.dat del /q XfoilData.dat\n");
    system("call XFoil < Xfoil.txt > XfoilLog.txt");
    
    // *** DATA FROM Xfoil ***
    
    double a, CL, CD, CDp, CM, Top_Xtr, Bot_Xtr;
    
    XfoilData(a, CL, CD, CDp, CM, Top_Xtr, Bot_Xtr);
    
    //*** OBJECTIV FUNCTION ***
    
    return CD/CL;
}

int Direction(ofstream &log, int i, int n, double *X, double F, double *dF, double *p)
{
    //Gradient count
           
    log << "dF = {";
           
    for(int x=0; x<n; x++)      //dF for every unknown   
    {
        dF[x] = Gradient(i, n, x, X, F); 
        log << dF[x];
            
        if(x != n-1) 
            log << ", ";
                
        else
            log << "}" << endl;
    }
        
    //Scale
        
    double Scale = 0;
    
    for(int x=0; x<n; x++)    
        Scale = Scale + dF[x]*dF[x];
    
    Scale = sqrt(Scale);
        
    //Direction vector p
        
    log << "p = {";
        
    for(int x=0; x<n; x++)      //dF for every unknown   
    {
        p[x] = -dF[x]/Scale;
        log << p[x];
            
        if(x != n-1) 
            log << ", ";
                
        else
            log << "}" << endl;
    }
        
    log << endl;
}

double Gradient(int i, int n, int x, double *X, double F)
{
    double eps = 0.001;
    double Xtemp, Feps, dF;
    
    //*** X init ***
    
    Xtemp = X[x];
    X[x] = X[x] + eps;
    
    //*** derevative ***
    
    X[x] = X[x] + eps;
    Feps = Xfoil(X);
    
    //*** previous X ***
    
    X[x] = Xtemp;
    
    return dF = (Feps - F) / eps;             
}

int LambdaSearch(ofstream &log, int i, int n, double Lamb, double AlfaLim, double *X, double &F, double *dF, double *p, data &aero)
{
    double c = 0.1;
    double Xtemp[3];
    double ThLamb, Th0, dTh0;
    
    for(int x=0; x<n; x++)
        Xtemp[x] = X[x]; 
            
    //Th0 estimation
        
    Th0 = F;
        
    //dTh0 estimation
        
    dTh0 = 0;
        
    for(int x=0; x<n; x++)
        dTh0 = dTh0 + dF[x]*p[x]; 
        
    log << "ARMIJO CONDITION: ThLamb <= Th0 + c*Lamb*dTh0" << endl << endl;
    log << "\tTh0 = " << Th0 << "\t\tdTh0 = " << dTh0 << endl << endl;
        
    for(int j=0; j<AlfaLim; j++)
    {
        //ThLamb estimation 
        
        for(int x=0; x<n; x++)
            X[x] = Xtemp[x] + Lamb*p[x];
        
        ThLamb = Xfoil(X);
        
        if(ThLamb <= Th0 + c*Lamb*dTh0)
        {
            log << "\t" << j << ": " << ThLamb << " <= " << Th0 + c*Lamb*dTh0 << " - APPROVED" << endl;
            F = ThLamb;
            
            XfoilData(aero.Alfa, aero.CL, aero.CD, aero.CDp, aero.CM, aero.Top_Xtr, aero.Bot_Xtr);
                
            break;
        }
            
        else
        {
            log << "\t" << j << ": " << ThLamb << " <= " << Th0 + c*Lamb*dTh0 << " - REJECTED" << endl;
                
            //*** QUADRIATIC INTERPOLATION ***
                
            double LambPrev = Lamb;
    
            Lamb = -0.5*dTh0*Lamb*Lamb / (ThLamb - Th0 - dTh0*Lamb); 
    
            if((LambPrev - Lamb < 1e-10) || (Lamb <= 0))
                Lamb = LambPrev*0.5;
                
            log << "\tNewLamb = " << Lamb << endl << endl;
        }
    }
}

int Stop(ofstream &log, int n, double Fprev, double F, double Sum_dF, double dF_Stop, double F_Stop)
{
    int stop = 1;
    double dif_F; 
    
    //*** SUM OF GRADIENT ***
            
    if(Sum_dF <= dF_Stop)
    {
        log << endl << "STOP CRYTERIUM:" << endl;
        log << "Sum_dF = " << Sum_dF << endl;
        stop = 0;
    }
    
    //*** OBJECTIVE FUNCTION DIFFERENCE ***
    
    dif_F = fabs(Fprev - F);
    
    if(dif_F <= F_Stop)
    {
        log << endl << "STOP CRYTERIUM:" << endl;
        log << "dif_F = " << dif_F << endl;
        stop = 0;
    }
    
    return stop;
}

int Results(ofstream &log, ofstream &res, int i, int n, double &Sum_dF, double *X, double F, double *dF, data aero)
{
    //sum of gradient count
    
    Sum_dF = 0;
        
    for(int x=1; x<n; x++)
        Sum_dF = Sum_dF + fabs(dF[x]);
    
    //print to file
        
    log << endl << "X = {" << X[0];
    res << i << "\t" << X[0] << "\t";
        
    for(int x=1; x<n; x++)  
    {    
        log << ", " << X[x];
        res << X[x] << "\t";
    }
               
    log << "}" << endl;
        
    log << "F = " << F << endl; 
        
    res << F << "\t" << 1/F << "\t" << Sum_dF << "\t\t";
    
    res << aero.Alfa << "\t" << aero.CL << "\t" << aero.CD << "\t" << aero.CDp << "\t"
        << aero.CM << "\t" << aero.Top_Xtr << "\t" << aero.Bot_Xtr << endl;
}

int XfoilData(double &a, double &CL, double &CD, double &CDp, double &CM, double &Top_Xtr, double &Bot_Xtr)
{
    // *** DATA FROM Xfoil ***
     
    string linia;
     
    ifstream in("XfoilData.dat");

    if(!in)
    {
        cout << "Data_out error!!!" << endl;
        return 0;
    }
       
    while(true)
    {
        getline(in, linia, '\n');
        
        if ("-" == linia.substr(1,1))  
            break;     
    }
   
    in >> a >> CL >> CD >> CDp >> CM >> Top_Xtr >> Bot_Xtr;
      
    in.close(); 
}
